import opensim as osim

def solveMocoInverse(model_fileName):
    inverse = osim.MocoInverse()
    modelProcessor = osim.ModelProcessor(model_fileName)
    modelProcessor.append(osim.ModOpAddExternalLoads('grf_walk.xml'))
    modelProcessor.append(osim.ModOpIgnoreTendonCompliance())
    modelProcessor.append(osim.ModOpIgnoreActivationDynamics())
    modelProcessor.append(osim.ModOpReplaceMusclesWithDeGrooteFregly2016())
    modelProcessor.append(osim.ModOpIgnorePassiveFiberForcesDGF())
    modelProcessor.append(osim.ModOpScaleActiveFiberForceCurveWidthDGF(1.5))
    modelProcessor.append(osim.ModOpAddReserves(1.0))
    inverse.setModel(modelProcessor)
    inverse.setKinematics(osim.TableProcessor('coordinates.sto'))
    inverse.set_initial_time(0.5)
    inverse.set_final_time(0.7) # 0.2 s only
    inverse.set_mesh_interval(0.02)
    inverse.set_kinematics_allow_extra_columns(True)
    solution = inverse.solve()
    solution.getMocoSolution().write('example3DWalking_MocoInverse_solution.sto')

# [info] Number of threads: 4

# %% A model with weak residual actuators
solveMocoInverse(model_fileName='subject_walk_armless.osim')
''' IPOPT output
    Number of nonzeros in equality constraint Jacobian...:    15708
    Number of nonzeros in inequality constraint Jacobian.:        0
    Number of nonzeros in Lagrangian Hessian.............:        0

    Total number of variables............................:     2058
                         variables with only lower bounds:        0
                    variables with lower and upper bounds:     1722
                         variables with only upper bounds:        0
    Total number of equality constraints.................:      378
    Total number of inequality constraints...............:        0
            inequality constraints with only lower bounds:        0
       inequality constraints with lower and upper bounds:        0
            inequality constraints with only upper bounds:        0

    iter    objective    inf_pr   inf_du lg(mu)  ||d||  lg(rg) alpha_du alpha_pr  ls
      34  1.6916193e+03 3.97e-04 7.91e-04  -4.8 3.10e-01    -  1.00e+00 9.78e-01h  1

    Number of Iterations....: 34

                                       (scaled)                 (unscaled)
    Objective...............:   1.6916192669677789e+03    1.6916192669677789e+03
    Dual infeasibility......:   7.9051671421254176e-04    7.9051671421254176e-04
    Constraint violation....:   3.9720671803422647e-04    3.9720671803422647e-04
    Variable bound violation:   0.0000000000000000e+00    0.0000000000000000e+00
    Complementarity.........:   1.5393805775190415e-04    1.5393805775190415e-04
    Overall NLP error.......:   7.9051671421254176e-04    7.9051671421254176e-04

    Number of objective function evaluations             = 40
    Number of objective gradient evaluations             = 35
    Number of equality constraint evaluations            = 40
    Number of inequality constraint evaluations          = 0
    Number of equality constraint Jacobian evaluations   = 35
    Number of inequality constraint Jacobian evaluations = 0
    Number of Lagrangian Hessian evaluations             = 0
    Total seconds in IPOPT                               = 97.270

    EXIT: Optimal Solution Found.
             nlp  :   t_proc      (avg)   t_wall      (avg)    n_eval
    callback_fun  |   5.00ms (142.86us)   7.02ms (200.70us)        35
           nlp_f  |  44.00ms (  1.10ms)  44.41ms (  1.11ms)        40
           nlp_g  |   1.66 s ( 41.43ms)   1.66 s ( 41.41ms)        40
      nlp_grad_f  | 416.00ms ( 11.24ms) 414.54ms ( 11.20ms)        37
       nlp_jac_g  |  94.14 s (  2.62 s)  94.14 s (  2.61 s)        36
           total  |  97.30 s ( 97.30 s)  97.30 s ( 97.30 s)         1

    Breakdown of objective (including weights):
      excitation_effort: 1691.62'''

# %% A model with strong residual actuators (1000)
solveMocoInverse(model_fileName='subject_walk_armless_strong.osim')
''' IPOPT output
    Number of nonzeros in equality constraint Jacobian...:    15183
    Number of nonzeros in inequality constraint Jacobian.:        0
    Number of nonzeros in Lagrangian Hessian.............:        0

    Total number of variables............................:     2058
                         variables with only lower bounds:        0
                    variables with lower and upper bounds:     1722
                         variables with only upper bounds:        0
    Total number of equality constraints.................:      378
    Total number of inequality constraints...............:        0
            inequality constraints with only lower bounds:        0
       inequality constraints with lower and upper bounds:        0
            inequality constraints with only upper bounds:        0

    iter    objective    inf_pr   inf_du lg(mu)  ||d||  lg(rg) alpha_du alpha_pr  ls
      27  2.0671467e-01 8.94e-04 9.32e-04  -4.8 5.08e-01    -  9.74e-01 1.00e+00f  1

    Number of Iterations....: 27

                                       (scaled)                 (unscaled)
    Objective...............:   2.0671467354857484e-01    2.0671467354857484e-01
    Dual infeasibility......:   9.3245392898411834e-04    9.3245392898411834e-04
    Constraint violation....:   8.9418398934526530e-05    8.9418398934526522e-04
    Variable bound violation:   0.0000000000000000e+00    0.0000000000000000e+00
    Complementarity.........:   4.0193309677586728e-05    4.0193309677586728e-05
    Overall NLP error.......:   9.3245392898411834e-04    9.3245392898411834e-04

    Number of objective function evaluations             = 30
    Number of objective gradient evaluations             = 28
    Number of equality constraint evaluations            = 30
    Number of inequality constraint evaluations          = 0
    Number of equality constraint Jacobian evaluations   = 28
    Number of inequality constraint Jacobian evaluations = 0
    Number of Lagrangian Hessian evaluations             = 0
    Total seconds in IPOPT                               = 66.683

    EXIT: Optimal Solution Found.
             nlp  :   t_proc      (avg)   t_wall      (avg)    n_eval
    callback_fun  |   3.00ms (107.14us)   5.20ms (185.70us)        28
           nlp_f  |  27.00ms (900.00us)  28.67ms (955.66us)        30
           nlp_g  | 925.00ms ( 30.83ms) 925.68ms ( 30.86ms)        30
      nlp_grad_f  | 268.00ms (  8.93ms) 273.86ms (  9.13ms)        30
       nlp_jac_g  |  64.97 s (  2.24 s)  64.96 s (  2.24 s)        29
           total  |  66.70 s ( 66.70 s)  66.70 s ( 66.70 s)         1

    Breakdown of objective (including weights):
      excitation_effort: 0.206715'''